// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/* written by John Doug Reynolds, September 2000 */

// $Id: gtibuild.cc,v 1.1.1.1 2004/06/07 13:08:53 eojero Exp $

#include "gtibuild.h"
#include <OGIPgti.h>
#include <STime.h>
#include <Xmm.h>

#include <STime.h>
#include <param.h>
#include <sstream>
#include <fstream>
#include <cmath>
#include <cstdio>

using namespace std;

class LineSeg {
public:
   virtual void appendTo( GtiList& ) const = 0;
   virtual void rejectIn( GtiList& ) const = 0;
};

class SimpleLineSeg : public LineSeg {
public:
   SimpleLineSeg( double min, double max );
   void appendTo( GtiList& ) const;
   void rejectIn( GtiList& ) const;
private:
   double _min, _max;
};

class DashedLineSeg : public LineSeg {
public:
   DashedLineSeg( double min, double max,
		  double seg1, double seg2, bool goodSegFirst );
   void appendTo( GtiList& ) const;
   void rejectIn( GtiList& ) const;
private:
   unsigned long _periods;
   double _start, _seg1, _seg2;
   bool _goodSegFirst;
};

double toDouble( const string &str )
{
   double val;
   if ( sscanf(str.c_str(),"%lf",&val) != 1 )
      errstr << "found '" << str << "' where a number was expected"
	     << Sas::error("badNumericValue");
   return val;
}

double toDouble( istringstream &str )
{
   string tok;
   if ( !(str >> tok) )
      Sas::error("unexpectedEOL","line ended before description complete");
   return toDouble(tok);
}

bool testWhich( char c, char y, char n )
{
   if ( c == y ) return true;
   if ( c != n )
      errstr << "found '" << c << "' where '"
	     << y << "' or '" << n << "' was expected"
	     << Sas::error("badOperator");
   return false;
}

void gtibuild()
{
  std::string filename = stringParameter("file");
  std::string tablename = stringParameter("table");
  
  struct GtiList gtiList;
  std::vector<LineSeg*> badSegs;
  
  // parse the GTI description file
  
  std::ifstream stream(filename.c_str());
  if ( !stream )
    errstr << "could not open: " << filename << Sas::error("FileDoesNotExist");
  
  std::string line;
  
  // first pass to identify any timeref
  double timeref = 0.;
  string strref;

  while(getline(stream, line)){
    std::istringstream s(line.c_str());
    std::string token;
    if(!(s >> token) || (token != "timeref")) continue;
    
    if(token == "timeref"){       
      if(s >> token)
	strref = token;
      continue;
    }
  }
  
  if(!strref.empty()){
    STime::ErrorHandling eh = STime::errorHandling;
    STime::errorHandling = STime::EXCEPTION;
    try {
      STime tz(strref);
      timeref = tz - Xmm::missionReferenceTime;
    }
    catch(STime::InvalidTime e){
      e.error();
      Sas::error("InvalidTime", "The string " + strref
		 + " is not a valid time identifier.");
    }
    STime::errorHandling = eh;
    Sas::message(Msg::App, Msg::Silent, "Reference time: "
		 + strref);
  }
  
  // Does not work. 
  //  stream.seekg(0);   
  // ... but clear has to be introduced to "rewind" the stream
  stream.clear();
  stream.close();
  stream.open(filename.c_str());
  while ( getline(stream,line) ) {
    istringstream s(line.c_str());
    string token;
    
    if ( !(s >> token) || (token[0] == '#') ||
	 token == "timeref") continue;
    
    double start = toDouble(token) + timeref;
    double stop = toDouble(s) + timeref;

    if ( !(s >> token) || token[0] == '#' ) token = "+";
    
    bool include = testWhich(token[0],'+','-');
    
    if ( token.size() > 1 && token[1] != '#' ) {
      if ( token.size() > 2 )
	errstr << "invalid operator: " << token
	       << Sas::error("badOperator");
      
      bool gsf = testWhich(token[1],'g','b');
      double seg1 = toDouble(s), seg2 = toDouble(s);
      
      if ( include )
	DashedLineSeg(start,stop,seg1,seg2,gsf).appendTo(gtiList);
      else
	badSegs.push_back( new DashedLineSeg(start,stop,seg1,seg2,gsf) );
    }
    else
      if ( include )
	SimpleLineSeg(start,stop).appendTo(gtiList);
      else
	badSegs.push_back( new SimpleLineSeg(start,stop) );
  }
  
  // all time is good if no good intervals were specified
  
  if ( gtiList.size() == 0 ) gtiList.push_back( Gti() );
  
  // mask out the bad intervals
  
  vector<LineSeg*>::const_iterator iter;
  for ( iter = badSegs.begin(); iter != badSegs.end(); ++iter )
    (*iter)->rejectIn(gtiList);
  
  // write out the GTI table
  
  struct OutOGIPgtiTable gtiTable(tablename);
  gtiTable << gtiList;
}

// --- LineSeg methods --------------------------------------------------------

SimpleLineSeg::SimpleLineSeg( double min, double max )
{
   if ( min < 0 || max < 0 )
      Sas::error("negativeTime","negative time values are not permitted");

   if ( max > 0 && max < min )
      { _min = max; _max = min; }
   else
      { _min = min; _max = max; }
}

void SimpleLineSeg::appendTo( struct GtiList &gtis ) const
{
   gtis.push_back( _max ? Gti(_min,_max) : Gti(_min) );
}

void SimpleLineSeg::rejectIn( struct GtiList &gtis ) const
{
   struct GtiList mask;
   if ( _min > 0 ) mask.push_back( Gti(0,_min) );
   if ( _max > 0 ) mask.push_back( Gti(_max) );
   gtis &= mask;
}

DashedLineSeg::DashedLineSeg( double min, double max,
			      double seg1, double seg2, bool goodSegFirst )
   : _start(max < min ? max : min), _seg1(seg1), _seg2(seg2),
     _goodSegFirst(goodSegFirst)
{
   if ( _start < 0 )
      Sas::error("negativeTime","negative time values are not permitted");
   if ( max == 0 )
      Sas::error("tooManyIntervals","periodic intervals must be bounded");
   if ( seg1 <= 0 || seg2 <= 0 )
      Sas::error("badSegment","periodic segments must be positive");

   _periods = (unsigned long) ceil( fabs(max - min) / (seg1 + seg2) );
}

void DashedLineSeg::appendTo( struct GtiList &gtis ) const
{
   double stride = _seg1 + _seg2;
   double width = _goodSegFirst ? _seg1 : _seg2;
   double start = _start + (_goodSegFirst ? 0 : _seg1);
   for ( unsigned long i = _periods; i--; start += stride )
      gtis.push_back( Gti(start,start+width) );
}

void DashedLineSeg::rejectIn( struct GtiList &gtis ) const
{
   struct GtiList mask;
   if ( _start > 0 ) mask.push_back( Gti(0,_start) );
   appendTo(mask);
   mask.push_back( Gti(_start + _periods * (_seg1+_seg2)) );
   gtis &= mask;
}
